/*
 * (c) 2017-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*****************************************************************************
 *                                                                           *
 * Driver for the Intel 8255x 10/100 Mbps Ethernet Controller Family         *
 * Network controller main header                                            *
 *                                                                           *
 *****************************************************************************/

#ifndef _E100_H
#define _E100_H

#include <io-pkt/iopkt_driver.h>
#include <stdio.h>
#include <errno.h>
#include <atomic.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/siginfo.h>
#include <sys/syspage.h>
#include <sys/neutrino.h>
#include <sys/mbuf.h>
#include <sys/syslog.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/if_ether.h>
#include <net/if_media.h>
#include <sys/io-pkt.h>
#include <sys/cache.h>
#include <sys/callout.h>
#include <hw/inout.h>
#if _KPDA_VERSION > 20180500
#include <netdrvr/mdi.h>
#include <netdrvr/eth.h>
#include <netdrvr/nicsupport.h>
#else
#include <drvr/mdi.h>
#include <drvr/eth.h>
#include <drvr/nicsupport.h>
#endif
#include <hw/nicinfo.h>
#include <hw/pci.h>
#include <nw_pci.h>
#include <hw/pci_devices.h>
#include <sys/device.h>
#include <quiesce.h>
#include <siglock.h>

#include <dev/mii/miivar.h>

#include <sys/slog.h>
#include <sys/slogcodes.h>

#include "bpfilter.h"
#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <net/if_vlanvar.h>

#include <net/ifdrvcom.h>
#include <sys/sockio.h>

#include <sys/malloc.h>
#include <device_qnx.h>

#if E100_DEBUG_LEVEL > 0
#	define E100_DEBUG(x) if ( e100->cfg.verbose >= E100_DEBUG_LEVEL ) { x; }
#	if E100_DEBUG_LEVEL > 1
#		define E100_DEBUG2(x) x
#	else
#		define E100_DEBUG2(x)
#	endif
# define RXID(__rx) (((void *)(__rx) - (void *)e100->rxs) / sizeof(struct rx))
#else
#	define E100_DEBUG(x)
#	define E100_DEBUG2(x)
#endif

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;
typedef u16 __le16;
typedef u32 __le32;
typedef u32 dma_addr_t;
#ifndef false
typedef int bool;
#define true (1==1)
#define false (1==0)
#endif
#define FULL 1
#define LITE 0

#define udelay(__usec) nanospin_ns((__usec)*1000);
#define msleep(__msec) delay((__msec))
#define msleep_interruptible msleep

#define cpu_to_le16(__val) ENDIAN_LE16((u16)(__val))
#define cpu_to_le32(__val) ENDIAN_LE32((u32)(__val))
#define le16_to_cpu cpu_to_le16
#define le32_to_cpu cpu_to_le32
#ifdef __X86__
#define iowrite8(__val, __port) ( \
	!e100->use_io ? (*(volatile uint8_t *)(__port) = (__val)) : \
	out8((uintptr_t)(__port), (__val)) \
)
#define iowrite16(__val, __port) ( \
	!e100->use_io ? (*(volatile uint16_t *)(__port) = cpu_to_le16(__val)): \
	out16((uintptr_t)(__port), cpu_to_le16(__val)) \
)
#define iowrite32(__val, __port) ( \
	!e100->use_io ? (*(volatile uint32_t *)(__port) = cpu_to_le32(__val)): \
	out32((uintptr_t)(__port), cpu_to_le32(__val)) \
)
#define ioread8(__port) ( \
	!e100->use_io ? *(volatile uint8_t *)(__port) : \
	in8((uintptr_t)(__port)) \
)
#define ioread16(__port) ( \
	!e100->use_io ? le16_to_cpu(*(volatile uint16_t *)(__port)) : \
	le16_to_cpu(in16((uintptr_t)(__port))) \
)
#define ioread32(__port) ( \
	!e100->use_io ? le32_to_cpu(*(volatile uint32_t *)(__port)) : \
	le32_to_cpu(in32((uintptr_t)(__port))) \
)
#else
#define iowrite8(__val, __port) (out8((uintptr_t)(__port), (__val)))
#define iowrite16(__val, __port) (out16((uintptr_t)(__port), cpu_to_le16(__val)))
#define iowrite32(__val, __port) (out32((uintptr_t)(__port), cpu_to_le32(__val)))
#define ioread8(__port) (in8((uintptr_t)(__port)))
#define ioread16(__port) (le16_to_cpu(in16((uintptr_t)(__port))))
#define ioread32(__port) (le32_to_cpu(in32((uintptr_t)(__port))))
#endif

#define dma_wmb() __cpu_membarrier()
#define dma_rmb() __cpu_membarrier()
#define FROM_POINTER_TO_INT(p, type) ( (type)(((uintptr_t)(p)) & ((1ULL << (sizeof(type) * 8)) - 1)) )

struct pci_device_id {
	u32 vendor, device;		/* Vendor and device ID or PCI_ANY_ID*/
	u32 driver_data;	/* Data private to the driver */
};

struct firmware_list {
	uint32_t start;
// 	uint32_t end;
	size_t size;
	char *name;
};

struct firmware {
	size_t size;
	const u8 *data;
};

enum mac {
	mac_82557_D100_A  = 0,
	mac_82557_D100_B  = 1,
	mac_82557_D100_C  = 2,
	mac_82558_D101_A4 = 4,
	mac_82558_D101_B0 = 5,
	mac_82559_D101M   = 8,
	mac_82559_D101S   = 9,
	mac_82550_D102    = 12,
	mac_82550_D102_C  = 13,
	mac_82551_E       = 14,
	mac_82551_F       = 15,
	mac_82551_10      = 16,
	mac_unknown       = 0xFF,
};

enum phy {
	phy_100a     = 0x000003E0,
	phy_100c     = 0x035002A8,
	phy_82555_tx = 0x015002A8,
	phy_nsc_tx   = 0x5C002000,
	phy_82562_et = 0x033002A8,
	phy_82562_em = 0x032002A8,
	phy_82562_ek = 0x031002A8,
	phy_82562_eh = 0x017002A8,
	phy_82552_v  = 0xd061004d,
	phy_unknown  = 0xFFFFFFFF,
};

/* CSR (Control/Status Registers) */
struct csr {
	struct {
		u8 status;
		u8 stat_ack;
		u8 cmd_lo;
		u8 cmd_hi;
		u32 gen_ptr;
	} scb;
	u32 port;
	u16 flash_ctrl;
	u8 eeprom_ctrl_lo;
	u8 eeprom_ctrl_hi;
	u32 mdi_ctrl;
	u32 rx_dma_count;
};

enum scb_status {
	rus_no_res       = 0x08,
	rus_ready        = 0x10,
	rus_mask         = 0x3C,
};

enum ru_state  {
	RU_SUSPENDED = 0,
	RU_RUNNING	 = 1,
	RU_UNINITIALIZED = -1,
};

enum scb_stat_ack {
	stat_ack_not_ours    = 0x00,
	stat_ack_sw_gen      = 0x04,
	stat_ack_mdi_done    = 0x08,
	stat_ack_rnr         = 0x10,
	stat_ack_cu_idle     = 0x20,
	stat_ack_frame_rx    = 0x40,
	stat_ack_cu_cmd_done = 0x80,
	stat_ack_not_present = 0xFF,
	stat_ack_rx = (stat_ack_sw_gen | stat_ack_rnr | stat_ack_frame_rx),
	stat_ack_tx = (stat_ack_cu_idle | stat_ack_cu_cmd_done),
};

enum scb_cmd_hi {
	irq_mask_none = 0x00,
	irq_mask_all  = 0x01,
	irq_sw_gen    = 0x02,
};

enum scb_cmd_lo {
	cuc_nop        = 0x00,
	ruc_start      = 0x01,
	ruc_resume     = 0x02,
	ruc_abort      = 0x04,
	ruc_load_base  = 0x06,
	cuc_start      = 0x10,
	cuc_resume     = 0x20,
	cuc_dump_addr  = 0x40,
	cuc_dump_stats = 0x50,
	cuc_load_base  = 0x60,
	cuc_dump_reset = 0x70,
};

enum cuc_dump {
	cuc_dump_complete       = 0x0000A005,
	cuc_dump_reset_complete = 0x0000A007,
};

enum port {
	software_reset  = 0x0000,
	selftest        = 0x0001,
	selective_reset = 0x0002,
};

enum eeprom_ctrl_lo {
	eesk = 0x01,
	eecs = 0x02,
	eedi = 0x04,
	eedo = 0x08,
};

enum mdi_ctrl {
	mdi_write = 0x04000000,
	mdi_read  = 0x08000000,
	mdi_ready = 0x10000000,
};

enum eeprom_op {
	op_write = 0x05,
	op_read  = 0x06,
	op_ewds  = 0x10,
	op_ewen  = 0x13,
};

enum eeprom_offsets {
	eeprom_cnfg_mdix  = 0x03,
	eeprom_map_cntr   = 0x05,
	eeprom_phy_iface  = 0x06,
	eeprom_id         = 0x0A,
	eeprom_config_asf = 0x0D,
	eeprom_smbus_addr = 0x90,
};

enum eeprom_cnfg_mdix {
	eeprom_mdix_enabled = 0x0080,
};

enum eeprom_phy_iface {
	NoSuchPhy = 0,
	I82553AB,
	I82553C,
	I82503,
	DP83840,
	S80C240,
	S80C24,
	II82555,
	DP83840A = 10,
};

enum eeprom_id {
	eeprom_id_wol = 0x0020,
};

enum eeprom_config_asf {
	eeprom_asf = 0x8000,
	eeprom_gcl = 0x4000,
};

enum cb_status {
	cb_complete = 0x8000,
	cb_ok       = 0x2000,
};

/**
 * cb_command - Command Block flags
 * @cb_tx_nc:  0: controller does CRC (normal),  1: CRC from skb memory
 */
enum cb_command {
	cb_nop    = 0x0000,
	cb_iaaddr = 0x0001,
	cb_config = 0x0002,
	cb_multi  = 0x0003,
	cb_tx     = 0x0004,
	cb_ucode  = 0x0005,
	cb_dump   = 0x0006,
	cb_tx_sf  = 0x0008,
	cb_tx_nc  = 0x0010,
	cb_cid    = 0x1f00,
	cb_i      = 0x2000,
	cb_s      = 0x4000,
	cb_el     = 0x8000,
};

struct rfd {
	__le16 status;
	__le16 command;
	__le32 link;
	__le32 rbd;
	__le16 actual_size;
	__le16 size;
};

struct rx {
	struct rx *next, *prev;
	struct mbuf *m;
	dma_addr_t dma_addr;
};

enum destroy_variants {
	DESTROY_ALL = -1,
	ATTACH_DONE,
	PCI_ENABLE_DONE,
	EVCNT_DONE,
	CACHE_INIT_DONE,
	MMAP_CSR_DONE,
	PCI_MASTER_DONE,
	MMAP_E100_DONE,
	ALLOC_CBS_DONE,
	ALLOC_RXS_DONE,
	INTER_ENTRY_DONE,
	IF_SETUP_DONE
};

#if defined(__BIGENDIAN__)
#define X(a,b)	b,a
#else
#define X(a,b)	a,b
#endif
struct config {
/*0*/	u8 X(byte_count:6, pad0:2);
/*1*/	u8 X(X(rx_fifo_limit:4, tx_fifo_limit:3), pad1:1);
/*2*/	u8 adaptive_ifs;
/*3*/	u8 X(X(X(X(mwi_enable:1, type_enable:1), read_align_enable:1),
	   term_write_cache_line:1), pad3:4);
/*4*/	u8 X(rx_dma_max_count:7, pad4:1);
/*5*/	u8 X(tx_dma_max_count:7, dma_max_count_enable:1);
/*6*/	u8 X(X(X(X(X(X(X(late_scb_update:1, direct_rx_dma:1),
	   tno_intr:1), cna_intr:1), standard_tcb:1), standard_stat_counter:1),
	   rx_save_overruns : 1), rx_save_bad_frames : 1);
/*7*/	u8 X(X(X(X(X(rx_discard_short_frames:1, tx_underrun_retry:2),
	   pad7:2), rx_extended_rfd:1), tx_two_frames_in_fifo:1),
	   tx_dynamic_tbd:1);
/*8*/	u8 X(X(mii_mode:1, pad8:6), csma_disabled:1);
/*9*/	u8 X(X(X(X(X(rx_tcpudp_checksum:1, pad9:3), vlan_arp_tco:1),
	   link_status_wake:1), arp_wake:1), mcmatch_wake:1);
/*10*/	u8 X(X(X(pad10:3, no_source_addr_insertion:1), preamble_length:2),
	   loopback:2);
/*11*/	u8 X(linear_priority:3, pad11:5);
/*12*/	u8 X(X(linear_priority_mode:1, pad12:3), ifs:4);
/*13*/	u8 ip_addr_lo;
/*14*/	u8 ip_addr_hi;
/*15*/	u8 X(X(X(X(X(X(X(promiscuous_mode:1, broadcast_disabled:1),
	   wait_after_win:1), pad15_1:1), ignore_ul_bit:1), crc_16_bit:1),
	   pad15_2:1), crs_or_cdt:1);
/*16*/	u8 fc_delay_lo;
/*17*/	u8 fc_delay_hi;
/*18*/	u8 X(X(X(X(X(rx_stripping:1, tx_padding:1), rx_crc_transfer:1),
	   rx_long_ok:1), fc_priority_threshold:3), pad18:1);
/*19*/	u8 X(X(X(X(X(X(X(addr_wake:1, magic_packet_disable:1),
	   fc_disable:1), fc_restop:1), fc_restart:1), fc_reject:1),
	   full_duplex_force:1), full_duplex_pin:1);
/*20*/	u8 X(X(X(pad20_1:5, fc_priority_location:1), multi_ia:1), pad20_2:1);
/*21*/	u8 X(X(pad21_1:3, multicast_all:1), pad21_2:4);
/*22*/	u8 X(X(rx_d102_mode:1, rx_vlan_drop:1), pad22:6);
	u8 pad_d102[9];
};

#define E100_MAX_MULTICAST_ADDRS	64
struct multi {
	__le16 count;
	u8 addr[E100_MAX_MULTICAST_ADDRS * ETH_MAC_LEN + 2/*pad*/];
};

/* Important: keep total struct u32-aligned */
#define UCODE_SIZE			134
#define E100_TX_FRAGMENTS	8
// #define E100_TX_FRAGMENTS	1
struct cb {
	__le16 status;
	__le16 command;
	__le32 link;
	union {
		u8 iaaddr[ETH_MAC_LEN];
		__le32 ucode[UCODE_SIZE];
		struct config config;
		struct multi multi;
		struct {
			u32 tbd_array;
			u16 tcb_byte_count;
			u8 threshold;
			u8 tbd_count;
			struct {
				__le32 buf_addr;
				__le16 size;
				u16 eol;
			} tbd[E100_TX_FRAGMENTS];
		} tcb;
		__le32 dump_buffer_addr;
	} u;
	struct cb *next, *prev;
	dma_addr_t dma_addr;
	void *data;
};

enum loopback {
	lb_none = 0, lb_mac = 1, lb_phy = 3,
};

struct stats {
	__le32 tx_good_frames, tx_max_collisions, tx_late_collisions,
		tx_underruns, tx_lost_crs, tx_deferred, tx_single_collisions,
		tx_multiple_collisions, tx_total_collisions;
	__le32 rx_good_frames, rx_crc_errors, rx_alignment_errors,
		rx_resource_errors, rx_overrun_errors, rx_cdt_errors,
		rx_short_frame_errors;
	__le32 fc_xmt_pause, fc_rcv_pause, fc_rcv_unsupported;
	__le16 xmt_tco_frames, rcv_tco_frames;
	__le32 complete;
};

struct mem {
	struct {
		u32 signature;
		u32 result;
	} selftest;
	struct stats stats;
	u8 dump_buf[596];
};

struct param_range {
	u32 min;
	u32 max;
	u32 count;
	u32 reap;
};

struct params {
	struct param_range rfds;
	struct param_range cbs;
};

#define ____cacheline_aligned __attribute__((__aligned__(NET_CACHELINE_SIZE)))
#define likely(x)	__builtin_expect(!!(x), 1)
#define unlikely(x)	__builtin_expect(!!(x), 0)

typedef struct _nic_nto_ext {
	struct ethercom 		ecom; /* common device */
	struct callout			hk_callout; /* house keeping */
	struct callout			mii_callout; /* link up/down */
	nic_config_t			cfg;
	nic_stats_t				stats;
	struct _iopkt_self		*iopkt;
	struct cache_ctrl		cachectl;
	void					*pci_dev_hdl;
	struct pci_dev_info		pci_info;
	unsigned short			pci_command;
	unsigned short			pci_status;

	u8 stat_ack;
	int prot_flags;
	int map_flags;
	int dying;
	volatile unsigned stopping;
	const struct sigevent	* (*isrp)(void *, int); ____cacheline_aligned
	u16 (*mdio_ctrl)(struct _nic_nto_ext *e100, u32 addr, u32 dir, u32 reg, u16 data);

	void					*sd_hook;
	int						kermask;
	int						use_io;
	int						eeprom_bad_csum_allow;
	int						force_rx_bug;
	int						extra_rnr;
#define E100_DEFAULT_HK_TIMEOUT 5
	unsigned				hk_timeout;
	__le16					tx_command;

	struct rx *rxs				____cacheline_aligned;
	struct rx *rx_to_use;
	struct rx *rx_to_clean;
	struct rfd blank_rfd;
	enum ru_state ru_running;

// 	intrspin_t cb_lock ____cacheline_aligned;
// 	intrspin_t cmd_lock;
	struct csr *csr ____cacheline_aligned;
	enum scb_cmd_lo cuc_cmd;
	unsigned int cbs_avail;
	struct cb *cbs;
	struct cb *cb_to_use;
	struct cb *cb_to_send;
	struct cb *cb_to_clean;
	
	enum {
		ich                = (1 << 0),
		promiscuous        = (1 << 1),
		multicast_all      = (1 << 2),
		wol_magic          = (1 << 3),
		ich_10h_workaround = (1 << 4),
	} flags					____cacheline_aligned;
	

	enum mac mac;
	enum phy phy;
	struct params params;
	int num_receive;
	int num_transmit;
	enum loopback loopback;

	struct mem *mem;
	dma_addr_t dma_addr;
	dma_addr_t cbs_dma_addr;
	volatile unsigned irq_sw_gen;
	
	/* TX descriptor and buffer tracking */
	struct evcnt		ev_txdrop ____cacheline_aligned;
	int					start_running;  // tx in progress

	mdi_t				*mdi;
	int					pkts_received;      // optimization to not probe phy
	int					force_advertise;
	int					tx_reap;
	int					bmstr;		/* Cpu to Pci physical memory translation */
	
	/* RX descriptor and buffer tracking */
	struct _iopkt_inter	inter ____cacheline_aligned;
	int					iid ____cacheline_aligned;

	struct mii_data		bsd_mii;	// for media devctls

	int					tmem_fd;
	char*				tmem;   // typed memory area

	u8 adaptive_ifs;
	u8 tx_threshold;
	u32 tx_frames;
	u32 tx_collisions;
	u32 tx_fc_pause;
	u32 tx_tco_frames;

	u32 rx_fc_pause;
	u32 rx_fc_unsupported;
	u32 rx_tco_frames;
	u32 rx_over_length_errors;

	u16 eeprom_wc;
	__le16 eeprom[256];
	struct firmware 	*fw;
} e100_dev_t;

struct e100_dev {
#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)
#define __devname__ (e100 && e100->ecom.ec_if.if_xname[0] ? e100->ecom.ec_if.if_xname : "fxp*")
	struct device		sc_dev;	/* common device */
	e100_dev_t			*sc_e100;
	char				filler[sizeof(e100_dev_t) + NET_CACHELINE_SIZE];
};

// main
int e100_init(struct ifnet *ifp);
void e100_stop(struct ifnet *ifp, int disable);


// mii
int e100_phy_supports_mii(e100_dev_t *e100);
int e100_findphy(e100_dev_t *e100);
int nic_force_advertise(e100_dev_t *e100);
u16 mdio_ctrl_hw(e100_dev_t *e100, u32 addr, u32 dir, u32 reg, u16 data);
void e100_mii_callout(void *arg);

// bsd_media
void bsd_mii_initmedia(e100_dev_t *e100);

// devctl
int e100_ioctl(struct ifnet * ifp, unsigned long cmd, caddr_t data);

// fw
int e100_load_ucode_wait(e100_dev_t *e100);

// hw
void e100_write_flush(e100_dev_t *e100);
int e100_eeprom_load(e100_dev_t *e100);
int e100_alloc_cbs(e100_dev_t *e100);
void e100_clean_cbs(e100_dev_t *e100, int full_clean);
void e100_enable_irq(e100_dev_t *e100);
void e100_disable_irq(e100_dev_t *e100);
int e100_hw_init(e100_dev_t *e100);
void e100_set_multicast_list(e100_dev_t *e100);
void e100_hw_dump_registers(e100_dev_t *e100, int desc);
int e100_exec_cb(e100_dev_t *e100, void *data, int (*cb_prepare)(e100_dev_t *, struct cb *));
int e100_configure(e100_dev_t *e100, struct cb *cb);
int e100_exec_cmd(e100_dev_t *e100, u8 cmd, dma_addr_t dma_addr);
void e100_hw_reset(e100_dev_t *e100);

// misc
int netif_running(e100_dev_t *e100);
int early_transmit(e100_dev_t *e100, struct nw_work_thread *wtp);
void setup_mmap_flags(e100_dev_t *e100);
int is_broadcast_mac(uint8_t *mac_addr);
int valid_mac_addr(char *mac_addr);
void setup_tmem(e100_dev_t *e100);
void e100_get_defaults(e100_dev_t *e100);
int pci_set_master(e100_dev_t *e100);
int pci_get_config(e100_dev_t *e100);
int pci_enable_device(e100_dev_t *e100);
void pci_disable_device(e100_dev_t *e100);
void fill_random_mac(char *mac_addr);
int e100_set_ringparam(e100_dev_t *e100, int num_receive, int num_transmit, int tx_reap);
void e100_adjust_adaptive_ifs(e100_dev_t *e100, int speed, int duplex);
void e100_update_stats(e100_dev_t *e100);

// interrupt
const struct sigevent *e100_isr_kermask(void *arg, int iid);
const struct sigevent *e100_isr(void *arg, int iid);
int e100_enable_interrupt_kermask(void *arg);
int e100_enable_interrupt(void *arg);
int e100_process_interrupt(void *arg, struct nw_work_thread *wtp);

// transmit
void e100_start(struct ifnet *ifp);
int e100_tx_clean(e100_dev_t *e100);

// receive
#ifndef RFD_ALIGNMENT_FUDGE
#define RFD_ALIGNMENT_FUDGE 0
#endif
#define E100_NAPI_WEIGHT 16
#define MAX_ETHER_LEN (ETHER_MAX_LEN + ETHER_VLAN_ENCAP_LEN)
#define RFD_BUF_LEN (sizeof(struct rfd) + MAX_ETHER_LEN)
#define	RFD_FROM_M(m) (struct rfd *)(mtod(m, struct rfd *) - 1)
int e100_rx_alloc_list(e100_dev_t *e100);
void e100_rx_clean_list(e100_dev_t *e100, int full_clean);
int e100_rx_init_list(e100_dev_t *e100, struct nw_work_thread *wtp);
void e100_start_receiver(e100_dev_t *e100, struct rx *rx);
void e100_rx_clean(e100_dev_t *e100, struct nw_work_thread *wtp, unsigned int *work_done, unsigned int work_to_do);

#endif
